# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from pyqrllib.pyqrllib import bin2hstr

from qrl.core.Wallet import Wallet, WalletEncryptionError, WalletVersionError, WalletDecryptionError
from qrl.core.misc import logger
from tests.misc.helper import set_qrl_dir

logger.initialize_default()


class TestWalletVer0(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestWalletVer0, self).__init__(*args, **kwargs)

    def test_init(self):
        with set_qrl_dir("wallet_ver0"):
            wallet = Wallet()
            self.assertIsNotNone(wallet)

    def test_read_ver0(self):
        with set_qrl_dir("wallet_ver0"):
            wallet = Wallet()
            self.assertEqual(1, len(wallet.address_items))
            self.assertEqual(wallet.version, 0)

            addr_item = wallet.address_items[0]
            self.assertFalse(addr_item.encrypted)

            self.assertEqual('Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             addr_item.qaddress)

            xmss0 = wallet.get_xmss_by_index(0)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0.address))

            xmss0b = wallet.get_xmss_by_address(xmss0.address)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0b.address))

    def test_read_secure_ver0(self):
        with set_qrl_dir("wallet_secure_ver0"):
            wallet = Wallet()
            self.assertEqual(1, len(wallet.address_items))
            self.assertEqual(wallet.version, 0)

            self.assertTrue(wallet.address_items[0].encrypted)

            wallet.decrypt_item_ver0(0, 'test1234')
            addr_item = wallet.address_items[0]

            self.assertEqual('Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             addr_item.qaddress)

            xmss0 = wallet.get_xmss_by_index(0)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0.address))

            xmss0b = wallet.get_xmss_by_address(xmss0.address)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0b.address))

    def test_read_wallet_ver0_saves_wallet_ver1(self):
        with set_qrl_dir("wallet_ver0"):
            wallet = Wallet()
            self.assertEqual(wallet.version, 0)

            wallet.version = 1
            wallet.save()

            wallet_reloaded = Wallet()
            self.assertEqual(wallet_reloaded.version, 1)

    def test_read_wallet_secure_ver0_saves_wallet_ver1_encrypted(self):
        with set_qrl_dir("wallet_secure_ver0"):
            wallet = Wallet()

            self.assertEqual(wallet.version, 0)

            # Wallet will not let you save an encrypted ver0 wallet as ver1. You have to decrypt it first.
            # This is because Qaddress is unencrypted in the ver1 wallet.
            with self.assertRaises(WalletVersionError):
                wallet.save()

            wallet.decrypt('test1234')
            wallet.encrypt('test1234')
            wallet.version = 1
            wallet.save()

            wallet_reloaded = Wallet()
            self.assertEqual(wallet_reloaded.version, 1)

            self.assertEqual(wallet_reloaded.address_items[0].qaddress,
                             'Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f')

            wallet_reloaded.decrypt('test1234')
            addr_item = wallet_reloaded.address_items[0]

            self.assertEqual('Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             addr_item.qaddress)

            xmss0 = wallet_reloaded.get_xmss_by_index(0)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0.address))

            xmss0b = wallet_reloaded.get_xmss_by_address(xmss0.address)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0b.address))


class TestWallet(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestWallet, self).__init__(*args, **kwargs)

    def test_read(self):
        with set_qrl_dir("wallet_ver1"):
            wallet = Wallet()
            self.assertEqual(1, len(wallet.address_items))
            self.assertEqual(wallet.version, 1)

            addr_item = wallet.address_items[0]
            self.assertFalse(addr_item.encrypted)

            self.assertEqual('Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             addr_item.qaddress)

            xmss0 = wallet.get_xmss_by_index(0)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0.address))

            xmss0b = wallet.get_xmss_by_address(xmss0.address)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0b.address))

    def test_read_secure(self):
        with set_qrl_dir("wallet_secure_ver1"):
            wallet = Wallet()
            self.assertEqual(1, len(wallet.address_items))
            self.assertEqual(wallet.version, 1)
            self.assertTrue(wallet.encrypted)

            addr_item = wallet.address_items[0]

            self.assertEqual('Q010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             addr_item.qaddress)

            wallet.decrypt_item(0, 'test1234')
            xmss0 = wallet.get_xmss_by_index(0)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0.address))

            xmss0b = wallet.get_xmss_by_address(xmss0.address)
            self.assertEqual('010400d9f1efe5b272e042dcc8ef690f0e90ca8b0b6edba0d26f81e7aff12a6754b21788169f7f',
                             bin2hstr(xmss0b.address))

    def test_create(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            self.assertEqual(0, len(wallet.address_items))

            xmss1 = wallet.add_new_address(4)
            self.assertEqual(1, len(wallet.address_items))

            xmss2 = wallet.get_xmss_by_index(0)

            self.assertEqual(xmss1.address, xmss2.address)
            self.assertEqual(xmss1.mnemonic, xmss2.mnemonic)

    def test_create_load(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            wallet.add_new_address(height=4)
            wallet.save()

            wallet_b = Wallet()
            self.assertEqual(1, len(wallet_b.address_items))

            self.assertEqual(wallet.address_items[0], wallet_b.address_items[0])

    def test_create_custom_hash_function(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            xmss = wallet.add_new_address(height=4, hash_function="sha2_256")
            self.assertEqual("sha2_256", xmss.hash_function)

    def test_create_custom_hash_function_load(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            xmss = wallet.add_new_address(height=4, hash_function="sha2_256")
            wallet.save()

            self.assertEqual("sha2_256", xmss.hash_function)

            wallet_reloaded = Wallet()
            self.assertEqual("sha2_256", wallet_reloaded.address_items[0].hashFunction)

    def test_integrity_behaviour(self):
        with set_qrl_dir("no_data"):
            TEST_KEY = 'mytestkey'
            wallet = Wallet()
            wallet.add_new_address(4)
            wallet.add_new_address(4)
            wallet.encrypt(TEST_KEY)

            # An encrypted wallet is not editable.
            with self.assertRaises(WalletEncryptionError):
                wallet.add_new_address(4)
            # You may not re-encrypt it.
            with self.assertRaises(WalletEncryptionError):
                wallet.encrypt(TEST_KEY)
            # You can save it.
            wallet.save()

            wallet.decrypt_item(1, TEST_KEY)
            # A partially encrypted wallet is not editable.
            with self.assertRaises(WalletEncryptionError):
                wallet.add_new_address(4)
            # You may not re-encrypt it.
            with self.assertRaises(WalletEncryptionError):
                wallet.encrypt(TEST_KEY)
            # You may not re-decrypt it.
            with self.assertRaises(WalletEncryptionError):
                wallet.decrypt(TEST_KEY)
            # You can't even save it.
            with self.assertRaises(WalletEncryptionError):
                wallet.save()

            wallet.decrypt_item(0, TEST_KEY)
            # A fully decrypted wallet is editable.
            wallet.add_new_address(4)
            # You may not re-decrypt it.
            with self.assertRaises(WalletEncryptionError):
                wallet.decrypt(TEST_KEY)
            # You can save it.
            wallet.save()

    def test_encrypt_wallet(self):
        # 2 unencrypted addresses. This should work.
        with set_qrl_dir("no_data"):
            TEST_KEY = 'mytestkey'
            wallet = Wallet()
            wallet.add_new_address(height=4)
            wallet.add_new_address(height=4)

            wallet.encrypt(TEST_KEY)

            self.assertTrue(wallet.encrypted)
            self.assertFalse(wallet.encrypted_partially)

    def test_decrypt_wallet(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            wallet.add_new_address(height=4)
            wallet.add_new_address(height=4)
            addresses = wallet.addresses

            TEST_KEY = 'mytestkey'
            wallet.encrypt(TEST_KEY)
            self.assertTrue(wallet.encrypted)

            wallet.decrypt(TEST_KEY)
            self.assertEqual(addresses, wallet.addresses)
            self.assertFalse(wallet.encrypted_partially)

    def test_decrypt_wallet_wrong(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            wallet.add_new_address(height=4)
            wallet.add_new_address(height=4)
            addresses = wallet.addresses

            TEST_KEY = 'mytestkey'
            wallet.encrypt(TEST_KEY)
            self.assertTrue(wallet.encrypted)

            TEST_KEY = 'mytestkey_is_wrong'
            with self.assertRaises(WalletDecryptionError):
                wallet.decrypt(TEST_KEY)
            self.assertEqual(addresses, wallet.addresses)
            self.assertFalse(wallet.encrypted_partially)

    def test_encrypt_save_reload(self):
        with set_qrl_dir("no_data"):
            wallet = Wallet()
            wallet.add_new_address(height=4)
            wallet.save()

            wallet_b = Wallet()

            self.assertEqual(1, len(wallet_b.address_items))
            self.assertEqual(wallet.address_items[0], wallet_b.address_items[0])

            TEST_KEY = 'mytestkey'

            wallet_b.encrypt(TEST_KEY)
            wallet_b.save()

            self.assertEqual(1, len(wallet_b.address_items))
            self.assertNotEqual(wallet.address_items[0], wallet_b.address_items[0])

            wallet_c = Wallet()
            self.assertEqual(1, len(wallet_c.address_items))
            self.assertTrue(wallet_c.address_items[0].encrypted)

            wallet_c.decrypt(TEST_KEY)
            self.assertFalse(wallet_c.address_items[0].encrypted)
            self.assertEqual(wallet.address_items[0], wallet_c.address_items[0])



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from pyqrllib import pyqrllib
from pyqrllib.pyqrllib import bin2hstr, getRandomSeed, str2bin, bin2mnemonic, mnemonic2bin  # noqa
from pyqrllib.pyqrllib import XmssFast, QRLDescriptor

hash_functions = {
    "shake128": pyqrllib.SHAKE_128,
    "shake256": pyqrllib.SHAKE_256,
    "sha2_256": pyqrllib.SHA2_256
}
hash_functions_reverse = {v: k for k, v in hash_functions.items()}


class XMSS(object):
    @staticmethod
    def from_extended_seed(extended_seed: bytes):
        if len(extended_seed) != 51:
            raise Exception('Extended seed should be 51 bytes long')

        descr = QRLDescriptor.fromBytes(extended_seed[0:3])
        if descr.getSignatureType() != pyqrllib.XMSS:
            raise Exception('Signature type nor supported')

        height = descr.getHeight()
        hash_function = descr.getHashFunction()
        tmp = XmssFast(extended_seed[3:], height, hash_function)
        return XMSS(tmp)

    @staticmethod
    def from_height(tree_height: int, hash_function="shake128"):
        if hash_function not in hash_functions:
            raise Exception("XMSS does not support this hash function!")

        seed = getRandomSeed(48, '')
        return XMSS(XmssFast(seed, tree_height, hash_functions[hash_function]))

    def __init__(self, _xmssfast):
        """
        :param
        tree_height: height of the tree to generate. number of OTS keypairs=2**tree_height
        :param _xmssfast:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.height
        4

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp._xmss.getSignatureSize()
        2308

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getPK() )
        '000200eb0372d56b886645e7c036b480be95ed97bc431b4e828befd4162bf432858df83191da3442686282b3d5160f25cf162a517fd2131f83fbf2698a58f9c46afc5d'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> len( tmp._xmss.getPK() )
        67

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getSK() ) == xmss_sk_expected1
        True

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getRoot() )
        'eb0372d56b886645e7c036b480be95ed97bc431b4e828befd4162bf432858df8'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getPKSeed() )
        '3191da3442686282b3d5160f25cf162a517fd2131f83fbf2698a58f9c46afc5d'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp._xmss.getIndex()
        0

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getSKSeed() )
        'eda313c95591a023a5b37f361c07a5753a92d3d0427459f34c7895d727d62816'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp._xmss.getSKPRF() )
        'b3aa2224eb9d823127d4f9f8a30fd7a1a02c6483d9c0f1fd41957b9ae4dfc63a'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr(tmp._xmss.getAddress())
        '00020096e5c065cf961565169e795803c1e60f521af7a3ea0326b42aa40c0e75390e5d8f4336de'
        """
        self._xmss = _xmssfast

    @property
    def hash_function(self) -> str:
        descr = self._xmss.getDescriptor()
        function_num = descr.getHashFunction()
        function_name = hash_functions_reverse[function_num]
        if not function_name:
            raise Exception("Could not reverse-lookup the hash function")

        return function_name

    @property
    def signature_type(self):
        descr = self._xmss.getDescriptor()
        answer = descr.getSignatureType()
        return answer

    @property
    def height(self):
        return self._xmss.getHeight()

    @property
    def _sk(self):
        """
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> len(tmp._sk)
        132

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr(tmp._sk) == xmss_sk_expected1
        True

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed2)
        >>> bin2hstr(tmp._sk) == xmss_sk_expected2
        True
        """
        return bytes(self._xmss.getSK())

    @property
    def pk(self):
        """
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr(tmp.pk) == xmss_pk_expected1
        True
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr(tmp.pk) == xmss_pk_expected2
        True
        """
        return bytes(self._xmss.getPK())

    @property
    def number_signatures(self) -> int:
        """
        Returns the number of signatures in the XMSS tree
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.number_signatures
        16
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.number_signatures
        16
        """
        return self._xmss.getNumberSignatures()

    @property
    def remaining_signatures(self):
        """
        Returns the number of remaining signatures in the XMSS tree
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.remaining_signatures
        16
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.remaining_signatures
        16
        """
        return self._xmss.getRemainingSignatures()

    @property
    def mnemonic(self) -> str:
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(hstr2bin(xmss_mnemonic_eseed1))
        >>> tmp.mnemonic == xmss_mnemonic_test1
        True
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(hstr2bin(xmss_mnemonic_eseed2))
        >>> tmp.mnemonic == xmss_mnemonic_test2
        True
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(mnemonic2bin(xmss_mnemonic_test1))
        >>> tmp.mnemonic == xmss_mnemonic_test1
        True
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(mnemonic2bin(xmss_mnemonic_test2))
        >>> tmp.mnemonic == xmss_mnemonic_test2
        True
        """
        return bin2mnemonic(self._xmss.getExtendedSeed())

    @property
    def address(self) -> bytes:
        return bytes(self._xmss.getAddress())

    @property
    def qaddress(self) -> str:
        return 'Q' + bin2hstr(self.address)

    @property
    def ots_index(self) -> int:
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.ots_index
        0
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.ots_index
        0
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> s = tmp.sign(str2bin("test"))
        >>> tmp.ots_index
        1
        """
        return self._xmss.getIndex()

    def set_ots_index(self, new_index):
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> xmss = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> xmss.set_ots_index(1)
        >>> xmss.ots_index
        1
        >>> from qrl.crypto.doctest_data import *
        >>> xmss = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> xmss.set_ots_index(10)
        >>> xmss.ots_index
        10
        """
        self._xmss.setIndex(new_index)

    @property
    def hexseed(self) -> str:
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> tmp.hexseed
        '000200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'

        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed2)
        >>> tmp.hexseed
        '000200010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101'
        """
        return bin2hstr(self._xmss.getExtendedSeed())

    @property
    def extended_seed(self):
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp.seed )
        '000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed2)
        >>> bin2hstr( tmp.extended_seed )
        '000200010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101'
        """
        return self._xmss.getExtendedSeed()

    @property
    def seed(self):
        """
        :return:
        :rtype:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr( tmp.seed )
        '000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed2)
        >>> bin2hstr( tmp.seed )
        '010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101010101'
        """
        return self._xmss.getSeed()

    def sign(self, message: bytes) -> bytes:
        """
        :param message:
        :return:

        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed1)
        >>> bin2hstr(tmp.sign(str2bin("test_message"))) == xmss_sign_expected1
        True
        >>> from qrl.crypto.doctest_data import *
        >>> tmp = XMSS.from_extended_seed(xmss_test_eseed2)
        >>> bin2hstr(tmp.sign(str2bin("test_message"))) == xmss_sign_expected2
        True
        """
        return bytes(self._xmss.sign(message))

    @staticmethod
    def get_height_from_sig_size(sig_size: int) -> int:
        min_size = 4 + 32 + 67 * 32

        if sig_size < min_size:
            raise Exception("Invalid Signature Size")

        if (sig_size - 4) % 32 != 0:
            raise Exception("Invalid Signature Size")

        height = (sig_size - min_size) // 32

        return height

    @staticmethod
    def validate_signature(signature, PK):
        height = XMSS.get_height_from_sig_size(len(signature))

        if height == 0 or 2 * int(bin2hstr(PK)[2:4]) != height:
            return False

        return True


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from collections import namedtuple
from unittest import TestCase

from mock import Mock, patch, MagicMock
from pyqrllib.pyqrllib import hstr2bin, bin2hstr

from qrl.core.misc import logger
from qrl.core.p2p import p2pPeerManager
from qrl.core.p2p.p2pfactory import P2PFactory, p2p_msg_priority
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.core.qrlnode import QRLNode
from qrl.generated import qrllegacy_pb2
from tests.misc.helper import replacement_getTime

logger.initialize_default()

Host = namedtuple('Host', ['host', 'port'])


# FIXME: These tests will soon be removed
class TestP2PProtocol(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestP2PProtocol, self).__init__(*args, **kwargs)

    def setUp(self):
        with patch('qrl.core.misc.ntp.getTime', new=replacement_getTime):
            self.channel = P2PProtocol()

        self.channel._observable = Mock()

        self.channel.factory = Mock(autospec=P2PFactory)
        self.channel.factory.p2p_msg_priority = p2p_msg_priority
        self.channel.factory._qrl_node = Mock(autospec=QRLNode)
        self.channel.factory._qrl_node.peer_manager = Mock(autospec=p2pPeerManager)
        self.channel.factory._qrl_node.peer_manager.is_banned = MagicMock(return_value=False)
        self.channel.factory._qrl_node.peer_manager.trusted_addresses = MagicMock(return_value=[])

        sample_peer_1 = Host('127.0.0.2', '9000')
        sample_host = Host('127.0.0.1', '9000')
        attrs = {'getPeer.return_value': sample_peer_1, 'getHost.return_value': sample_host}
        self.channel.transport = Mock(**attrs)

    def tearDown(self):
        del self.channel

    def test_addr_remote_works(self):
        """
        This is more to ensure that the transport is mocked correctly.
        """
        self.assertEqual('127.0.0.2:9000', self.channel.peer.full_address)

    def test_addr_local_works(self):
        """
        This is more to ensure that the transport is mocked correctly.
        """
        self.assertEqual('127.0.0.1:9000', self.channel.host.full_address)

    @patch('qrl.core.misc.ntp.getTime')
    def test_connectionMade_behavior(self, getTime):
        """
        When connectionMade, the Managers (Observers) must be informed once, and peer list, version request messages
        must be in the outgoing message queue.
        ntp.getTime() is patched everywhere, not just in p2pprotocol, because OutgoingMessage uses it too.
        """
        self.channel.factory.get_connected_peer_addrs.return_value = {'127.0.0.2:9000'}

        getTime.return_value = 1525078652.9991353
        self.channel.connectionMade()

        # Managers were notified
        self.channel.p2pchain_manager.new_channel.assert_called_once_with(self.channel)
        self.channel.peer_manager.new_channel.assert_called_once_with(self.channel)
        self.channel.tx_manager.new_channel.assert_called_once_with(self.channel)

        # send_peer_list and send_version_request messages should be in the outgoing queue.
        self.assertEqual(self.channel.outgoing_queue.unfinished_tasks, 2)

    def test_connectionLost_behavior(self):
        """
        When connectionLost, peer_manager (an Observer) is notified. (why not other Observers?)
        """
        self.channel.connectionLost()
        self.channel.peer_manager.remove_channel.assert_called_once_with(self.channel)

    @patch('qrl.core.misc.ntp.getTime')
    def test_dataReceived_normal_message(self, getTime):
        """
        Ensure that dataReceived works as expected with a normal message.
        """
        getTime.return_value = 1525078652.9991353
        data = b'\x00\x00\x00\x80\x08\x01"|\n\x0e66.175.217.203\n\x0e138.195.214.85\n\r35.177.72.178\n\x0e173.249.22.240\n\x0c2.238.131.20\n\r77.64.144.198\n\r34.208.138.15\n\x0f144.202.107.148\x00\x00\x00\x00'  # noqa
        self.channel.dataReceived(data)

        # Twisted transport should have received acknowledgement message to send out
        acknowledgement_bytes = b'\x00\x00\x00\x08\x08\x13\xaa\x01\x03\x08\x84\x01'
        self.channel.transport.write.assert_called_once_with(acknowledgement_bytes)

    @patch('qrl.core.misc.ntp.getTime')
    @patch('qrl.core.p2p.p2pprotocol.logger', autospec=True)
    @patch('qrl.core.p2p.p2pprotocol.config.dev', autospec=True)
    def test_dataReceived_too_big(self, config_dev, logger, getTime):
        """
        Normally the buffer size upper limit is 10MB. But we're going to patch it smaller here.
        """
        config_dev.max_bytes_out = 10
        config_dev.trust_min_msgcount = 10
        getTime.return_value = 1525078652.9991353
        acknowledgement_bytes = b'\x00\x00\x00\x08\x08\x13\xaa\x01\x03\x08\x88\x01'
        self.channel._buffer = 10 * acknowledgement_bytes
        self.channel.dataReceived(acknowledgement_bytes)
        self.channel.transport.loseConnection.assert_called()

    @patch('qrl.core.misc.ntp.getTime')
    def test_dataReceived_spam_ban_peer(self, getTime):
        getTime.return_value = 1525078652.9991353
        self.channel.rate_limit = 2
        acknowledgement_bytes = b'\x00\x00\x00\x08\x08\x13\xaa\x01\x03\x08\x88\x01'
        self.channel._buffer = 10 * acknowledgement_bytes
        self.channel.dataReceived(acknowledgement_bytes)
        self.channel.peer_manager.ban_channel.assert_called_with(self.channel)

    @patch('qrl.core.misc.ntp.getTime')
    def test_send_version_request(self, getTime):
        getTime.return_value = 1525078652.9991353
        version_request = b'\x00\x00\x00\x02\x1a\x00'
        self.channel.send_version_request()
        self.channel.transport.write.assert_called_with(version_request)

    @patch('qrl.core.misc.ntp.getTime')
    def test_send_sync(self, getTime):
        getTime.return_value = 1525078652.9991353
        self.channel.send_sync(synced=True)
        synced = b'\x00\x00\x00\r\x08\x10\x92\x01\x08\n\x06Synced'
        self.channel.transport.write.assert_called_with(synced)

        self.channel.send_sync(synced=False)
        unsynced = b'\x00\x00\x00\x05\x08\x10\x92\x01\x00'
        self.channel.transport.write.assert_called_with(unsynced)

    @patch('qrl.core.misc.ntp.getTime')
    def test_send_fetch_block(self, getTime):
        getTime.return_value = 1525078652.9991353
        block_request = b'\x00\x00\x00\x06\x08\x06B\x02\x08\x01'
        self.channel.send_fetch_block(1)
        self.channel.transport.write.assert_called_with(block_request)

    @patch('qrl.core.misc.ntp.getTime')
    def test_get_headerhash_list(self, getTime):
        getTime.return_value = 1525078652.9991353
        get_headerhash_request = b'\x00\x00\x00\x05\x08\x12\xa2\x01\x00'
        self.channel.send_get_headerhash_list(1)
        self.channel.transport.write.assert_called_with(get_headerhash_request)

    def test_parse_buffer_works(self):
        self.channel._buffer = bytes(hstr2bin('000000191a170a0776657273696f6e120c67656e657369735f68617368' +
                                              '000000191a170a0776657273696f6e120c67656e657369735f68617368'))
        messages = self.channel._parse_buffer([0])
        self.assertEqual(2, len(list(messages)))

    @patch('qrl.core.p2p.p2pprotocol.logger', autospec=True)
    def test_parse_buffer_invalid_data(self, logger):
        self.channel._buffer = bytes(hstr2bin('0000000000000000000000000000000000000000000000000000000000' +
                                              '1111111111111111111111111111111111111111111111111111111111'))

        messages = self.channel._parse_buffer([0])
        messages_list = list(messages)
        self.assertEqual(0, len(messages_list))
        logger.warning.assert_called_with("Problem parsing message. Banning+Dropping connection")

    def test_wrap_message_works(self):
        veData = qrllegacy_pb2.VEData(version="version", genesis_prev_hash=b'genesis_hash')
        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE, veData=veData)
        self.assertEqual('000000191a170a0776657273696f6e120c67656e657369735f68617368',
                         bin2hstr(P2PProtocol._wrap_message(msg)))



# coding=utf-8
from __future__ import print_function
import simplejson as json
import yaml
import sys

from pyqrllib.pyqrllib import hstr2bin

from qrl.generated import qrl_pb2
from qrl.core import config
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.Block import Block
from qrl.crypto.xmss import XMSS


def create_tx(addrs_to, amounts, signing_xmss, nonce):
    tx = TransferTransaction.create(addrs_to=addrs_to,
                                    amounts=amounts,
                                    fee=0,
                                    xmss_pk=signing_xmss.pk)
    tx.sign(signing_xmss)
    tx._data.nonce = nonce
    return tx


def get_migration_transactions(signing_xmss, filename):
    transactions = []

    with open(filename, 'r') as f:
        json_data = json.load(f)

    count = 1
    addrs_to = []
    amounts = []
    output_limit = config.dev.transaction_multi_output_limit

    for addr in json_data:
        try:
            addrs_to.append(bytes(hstr2bin(addr[1:])))
        except:  # noqa
            print("Invalid Address ", addr)
            raise Exception
        amounts.append(json_data[addr])

        count += 1
        if count % output_limit == 0:
            transactions.append(create_tx(addrs_to, amounts, signing_xmss, count // output_limit))

            addrs_to = []
            amounts = []

    if addrs_to:
        transactions.append(create_tx(addrs_to, amounts, signing_xmss, (count // output_limit) + 1))

    return transactions


def main():
    if len(sys.argv) > 2:
        print("Unexpected arguments")
        sys.exit(0)
    elif len(sys.argv) == 1:
        print("Missing Filename")
        sys.exit(0)

    filename = sys.argv[1]

    if sys.version_info.major > 2:
        seed = bytes(hstr2bin(input('Enter extended hexseed: ')))
    else:
        seed = bytes(hstr2bin(raw_input('Enter extended hexseed: ')))  # noqa

    dist_xmss = XMSS.from_extended_seed(seed)

    transactions = get_migration_transactions(signing_xmss=dist_xmss, filename=filename)

    block = Block.create(block_number=0,
                         prev_headerhash=config.user.genesis_prev_headerhash,
                         prev_timestamp=config.user.genesis_timestamp,
                         transactions=transactions,
                         miner_address=dist_xmss.address)

    block.set_nonces(0, 0)

    block._data.genesis_balance.MergeFrom([qrl_pb2.GenesisBalance(address=config.dev.coinbase_address,
                                                                  balance=105000000000000000)])

    with open('genesis.yml', 'w') as f:
        yaml.dump(json.loads(block.to_json()), f)


if __name__ == '__main__':
    main()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import qrl.generated.qrldebug_pb2 as qrldebug__pb2


class DebugAPIStub(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Debug API used for debugging
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetFullState = channel.unary_unary(
        '/qrl.DebugAPI/GetFullState',
        request_serializer=qrldebug__pb2.GetFullStateReq.SerializeToString,
        response_deserializer=qrldebug__pb2.GetFullStateResp.FromString,
        )


class DebugAPIServicer(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Debug API used for debugging
  """

  def GetFullState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_DebugAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetFullState': grpc.unary_unary_rpc_method_handler(
          servicer.GetFullState,
          request_deserializer=qrldebug__pb2.GetFullStateReq.FromString,
          response_serializer=qrldebug__pb2.GetFullStateResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.DebugAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from unittest import TestCase

from mock import patch, Mock

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.MessageTransaction import MessageTransaction
from tests.misc.helper import get_alice_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.Transaction._revert_state_changes_for_PK')
@patch('qrl.core.txs.Transaction.Transaction._apply_state_changes_for_PK')
@patch('qrl.core.txs.Transaction.logger')
class TestMessageTransactionStateChanges(TestCase):
    def setUp(self):
        self.alice = get_alice_xmss()

        self.params = {
            "message_hash": b'Test Message',
            "fee": 1,
            "xmss_pk": self.alice.pk
        }
        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def generate_addresses_state(self, tx):
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[],
                                     balance=100),
        }
        return addresses_state

    def test_apply_state_changes(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)

        m_apply_state_PK.assert_called_once()

    def test_apply_state_changes_empty_addresses_state(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = {}
        tx.apply_state_changes(addresses_state)

        self.assertEqual({}, addresses_state)
        m_apply_state_PK.assert_called_once()

    def test_revert_state_changes(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)

        m_revert_state_PK.assert_called_once()

    def test_revert_state_changes_empty_addresses_state(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = {}
        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual({}, addresses_state)
        m_revert_state_PK.assert_called_once()

    def test_validate_tx(self, m_logger, m_apply_state_PK, m_revert_state_PK):
        tx = MessageTransaction.create(**self.params)
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from qrl.core.misc import logger
from tests.blockchain.MockedBlockchain import MockedBlockchain

logger.initialize_default()


class TestBlockchain(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBlockchain, self).__init__(*args, **kwargs)

    def test_forking_basic(self):
        """
        TODO: Cyyber review
        - Create a blockchain of 10 blocks
        - Add a new block 10b (linking to 9)
        - Check that it did not switch
        """
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            # Fork at node 9
            block_9 = mock_blockchain.qrlnode.get_block_from_index(9)
            block_10 = mock_blockchain.qrlnode.get_block_last()
            block_10b = mock_blockchain.create_block(block_9.headerhash)

            self.assertEqual(block_9.headerhash, block_10.prev_headerhash)
            self.assertEqual(block_9.headerhash, block_10b.prev_headerhash)

            # Add a new block pointing to 9
            mock_blockchain.add_block(block_10b)

            last_block = mock_blockchain.qrlnode.get_block_last()
            block_index_10 = mock_blockchain.qrlnode.get_block_from_index(10)

            self.assertEqual(last_block._data, block_index_10._data)
            self.assertEqual(last_block._data, block_10._data)
            self.assertNotEquals(last_block._data, block_10b._data)

    def test_forking_basic2(self):
        """
        TODO: Cyyber review
        - Create a blockchain of 10 blocks
        - Add a new blocks 10b, 11b (linking from 9)
        - Check that it did switch
        """
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            # Fork at node 9
            block_9 = mock_blockchain.qrlnode.get_block_from_index(9)

            # Fork with 2 blocks
            block_10b = mock_blockchain.create_block(block_9.headerhash)
            mock_blockchain.add_block(block_10b)
            block_11b = mock_blockchain.create_block(block_10b.headerhash)
            mock_blockchain.add_block(block_11b)

            last_block = mock_blockchain.qrlnode.get_block_last()
            print(last_block.block_number)
            block_index_10 = mock_blockchain.qrlnode.get_block_from_index(10)
            block_index_11 = mock_blockchain.qrlnode.get_block_from_index(11)

            self.assertEqual(block_10b._data, block_index_10._data)
            self.assertEqual(last_block._data, block_index_11._data)
            self.assertEqual(last_block._data, block_11b._data)

    def test_forking_basic3(self):
        """
        TODO: Cyyber review
        - Create a blockchain of 10 blocks
        - Add a new blocks 10b, 11b (linking from 9)
        - Check that it did switch
        - Add a new blocks 11, 12 (linking from 10)
        - Check that it did switch again
        """
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            # Fork at node 9
            block_9 = mock_blockchain.qrlnode.get_block_from_index(9)
            block_10 = mock_blockchain.qrlnode.get_block_last()

            # Fork with 2 blocks (b)
            block_10b = mock_blockchain.create_block(block_9.headerhash)
            mock_blockchain.add_block(block_10b)
            block_11b = mock_blockchain.create_block(block_10b.headerhash)
            mock_blockchain.add_block(block_11b)

            # Check
            last_block = mock_blockchain.qrlnode.get_block_last()
            block_index_10 = mock_blockchain.qrlnode.get_block_from_index(10)
            block_index_11 = mock_blockchain.qrlnode.get_block_from_index(11)
            self.assertEqual(block_10b._data, block_index_10._data)
            self.assertEqual(last_block._data, block_index_11._data)
            self.assertEqual(last_block._data, block_11b._data)

            # Add back to the original chain (a)
            block_11 = mock_blockchain.create_block(block_10.headerhash)
            mock_blockchain.add_block(block_11)
            block_12 = mock_blockchain.create_block(block_11.headerhash)
            mock_blockchain.add_block(block_12)

            # Check
            last_block = mock_blockchain.qrlnode.get_block_last()
            block_index_10 = mock_blockchain.qrlnode.get_block_from_index(10)
            block_index_11 = mock_blockchain.qrlnode.get_block_from_index(11)
            block_index_12 = mock_blockchain.qrlnode.get_block_from_index(12)
            self.assertEqual(block_10._data, block_index_10._data)
            self.assertEqual(block_11._data, block_index_11._data)
            self.assertEqual(last_block._data, block_index_12._data)
            self.assertEqual(last_block._data, block_12._data)
